@@ -23,19 +23,15 @@
 #ifndef _GTK_HTML_EMBEDDED_H
 #define _GTK_HTML_EMBEDDED_H
 
-#include <gtk/gtk.h>
+#include <gtk/gtkbin.h>
 #include <libgnome/gnome-defs.h>
 
+#include "gtkhtml-types.h"
+
 #define GTK_HTML_EMBEDDED(obj)         GTK_CHECK_CAST (obj, gtk_html_embedded_get_type (), GtkHTMLEmbedded)
 #define GTK_HTML_EMBEDDED_CLASS(klass) GTK_CHECK_CLASS_CAST (klass, gtk_html_embedded_get_type (), GtkHTMLEmbeddedClass)
 #define GTK_IS_HTML_EMBEDDED(obj)      GTK_CHECK_TYPE (obj, gtk_html_embedded_get_type ())
 
-
-typedef struct _GtkHTMLEmbedded      GtkHTMLEmbedded;
-typedef struct _GtkHTMLEmbeddedClass GtkHTMLEmbeddedClass;
-
-typedef struct _GtkHTMLEmbeddedPrivate GtkHTMLEmbeddedPrivate;
-
 struct _GtkHTMLEmbedded {
 	GtkBin bin;
 
@@ -43,6 +39,7 @@ struct _GtkHTMLEmbedded {
 	char *classid;
 	char *name;
         char *type;
+        char *data;
 
 	/* parameters to class */
 	int width, height;
@@ -63,7 +60,7 @@ struct _GtkHTMLEmbeddedClass {
    themselves for a form */
 
 guint		gtk_html_embedded_get_type	(void);
-GtkWidget	*gtk_html_embedded_new (char *classid, char *name, char *type, int width, int height);
+GtkWidget	*gtk_html_embedded_new (char *classid, char *name, char *type, char *data, int width, int height);
 
 void gtk_html_embedded_set_parameter (GtkHTMLEmbedded *ge, char *param, char *value);
 char *gtk_html_embedded_get_parameter (GtkHTMLEmbedded *ge, char *param);
