@@ -22,8 +22,10 @@
 #ifndef _GTKHTML_PRIVATE_H
 #define _GTKHTML_PRIVATE_H
 
-#include "gtkhtml.h"
-#include "gtkhtmldebug.h"
+#include <config.h>
+
+#include "gtkhtml-types.h"
+#include "htmltypes.h"
 
 
 struct _GtkHTMLPrivate {
@@ -35,20 +37,20 @@ struct _GtkHTMLPrivate {
 	GtkHTMLParagraphAlignment paragraph_alignment;
 	GtkHTMLFontStyle insertion_font_style;
 
-#ifdef GTKHTML_HAVE_PSPELL
+	gint last_selection_type;
+	gchar *content_type;
+
+#ifdef GTKHTML_HAVE_GCONF
 	guint set_font_id;
 #endif
-#ifdef GTKHTML_USE_XIM
+#ifdef GTK_HTML_USE_XIM
 	GdkICAttr *ic_attr;
 	GdkIC *ic;
 #endif
 };
 
-void  gtk_html_private_calc_scrollbars  (GtkHTML *html);
+void  gtk_html_private_calc_scrollbars  (GtkHTML            *html);
+void  gtk_html_editor_event_command     (GtkHTML            *html,
+					 GtkHTMLCommandType  com_type);
 
 #endif /* _GTKHTML_PRIVATE_H */
-
-
-
-
-
