@@ -22,6 +22,7 @@
 #include <config.h>
 
 #include <gdk/gdkkeysyms.h>
+#include <gdk/gdkprivate.h>
 #include <gtk/gtk.h>
 #include <string.h>
 #include <gal/widgets/e-unicode.h>
@@ -29,15 +30,11 @@
 #include <gnome.h>
 
 #include "htmlengine-edit-clueflowstyle.h"
-#include "htmlengine-edit-copy.h"
-#include "htmlengine-edit-cut.h"
-#include "htmlengine-edit-delete.h"
+#include "htmlengine-edit-cut-and-paste.h"
 #include "htmlengine-edit-fontstyle.h"
-#include "htmlengine-edit-insert.h"
 #include "htmlengine-edit-rule.h"
 #include "htmlengine-edit-movement.h"
 #include "htmlengine-edit-cursor.h"
-#include "htmlengine-edit-paste.h"
 #include "htmlengine-edit-text.h"
 #include "htmlengine-edit.h"
 #include "htmlengine-print.h"
@@ -268,14 +265,16 @@ idle_handler (gpointer data)
 	html = GTK_HTML (data);
 	engine = html->engine;
 
-	html_engine_make_cursor_visible (engine);
+	if (html->engine->thaw_idle_id == 0)
+		html_engine_make_cursor_visible (engine);
 
 	gtk_adjustment_set_value (GTK_LAYOUT (html)->hadjustment, (gfloat) engine->x_offset);
 	gtk_adjustment_set_value (GTK_LAYOUT (html)->vadjustment, (gfloat) engine->y_offset);
 
 	gtk_html_private_calc_scrollbars (html);
 
-	html_engine_flush_draw_queue (engine);
+	if (html->engine->thaw_idle_id == 0)
+		html_engine_flush_draw_queue (engine);
 
  	html->priv->idle_handler_id = 0;
 	return FALSE;
@@ -635,15 +634,15 @@ key_press_event (GtkWidget *widget,
 	    && ! (event->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK))
 	    && event->length > 0) {
 		gchar *str;
-		html_engine_delete_selection (html->engine, TRUE);
+
 		/* printf ("event length: %d s[0]: %d string: '%s'\n", event->length, event->string [0], event->string); */
 		str = e_utf8_from_gtk_event_key (widget, event->keyval, event->string);
 		/* printf ("len: %d str: %s\n", str ? unicode_strlen (str, -1) : -1, str); */
 		if (str)
-			html_engine_insert (html->engine, str, unicode_strlen (str, -1));
+			html_engine_paste_text (html->engine, str, unicode_strlen (str, -1));
 		else if (event->length == 1 && event->string
 			 && ((guchar)event->string [0]) > 0x20 && ((guchar)event->string [0]) < 0x80)
-			html_engine_insert (html->engine, event->string, 1);
+			html_engine_paste_text (html->engine, event->string, 1);
 		g_free (str);
 		retval = TRUE;
 	}
@@ -652,7 +651,6 @@ key_press_event (GtkWidget *widget,
 		html_engine_reset_blinking_cursor (html->engine);
 
 	if (retval) {
-		queue_draw (html);
 		if (position != html->engine->cursor->position)
 			update_styles (html);
 	}
@@ -978,7 +976,12 @@ button_press_event (GtkWidget *widget,
 			break;
 		case 2:
 			if (html_engine_get_editable (engine)) {
-				gtk_html_request_paste (widget, 0, event->time);
+				html_engine_disable_selection (html->engine);
+				html_engine_jump_at (engine,
+						     x + engine->x_offset,
+						     y + engine->y_offset);
+				update_styles (html);
+				gtk_html_request_paste (html, 0, event->time);
 				return TRUE;
 			}
 			break;
@@ -1042,7 +1045,8 @@ button_release_event (GtkWidget *widget,
 	}
 
 	if (html->in_selection) {
-		gtk_selection_owner_set (widget, GDK_SELECTION_PRIMARY, event->time);
+		if (html_engine_get_editable (html->engine))
+			gtk_html_copy (html);
 		html->in_selection = FALSE;
 		update_styles (html);
 	}
@@ -1122,8 +1126,10 @@ selection_get (GtkWidget        *widget,
 	g_return_if_fail (GTK_IS_HTML (widget));
 	
 	html = GTK_HTML (widget);
-	selection_string = html_engine_get_selection_string (html->engine);
-	
+	selection_string = html->engine->clipboard
+		? html_object_get_selection_string (html->engine->clipboard)
+		: html_engine_get_selection_string (html->engine);
+
 	/*
 	 * FIXME we should make e_utf8_to/from_string_target and 
 	 * e_utf8_to/from_compound_string in e-unicode.c but I'll wait until
@@ -1188,7 +1194,13 @@ selection_received (GtkWidget *widget,
 	g_return_if_fail (selection_data != NULL);
 	
 	printf("got selection from system\n");
-	
+
+	if (html_engine_get_editable (GTK_HTML (widget)->engine)
+	    && widget->window == gdk_selection_owner_get (selection_data->selection) && GTK_HTML (widget)->engine->clipboard) {
+		html_engine_paste (GTK_HTML (widget)->engine);
+		return;
+	}
+
 	/* **** IMPORTANT **** Check to see if retrieval succeeded  */
 	/* If we have more selection types we can ask for, try the next one,
 	   until there are none left */
@@ -1196,7 +1208,7 @@ selection_received (GtkWidget *widget,
 		gint type = GTK_HTML (widget)->priv->last_selection_type;
 		
 		/* now, try again with next selection type */
-		if (!gtk_html_request_paste(widget, type + 1, time))
+		if (!gtk_html_request_paste (GTK_HTML (widget), type + 1, time))
 			g_warning ("Selection retrieval failed\n");
 		return;
 	}
@@ -1206,19 +1218,14 @@ selection_received (GtkWidget *widget,
 	    && (selection_data->type != GDK_SELECTION_TYPE_STRING)
 	    && (selection_data->type != gdk_atom_intern ("UTF-8", FALSE))) {
 		g_warning ("Selection \"STRING\" was not returned as strings!\n");
-		return;
-	}
-	
-	if (selection_data->length > 0) {
+	} else if (selection_data->length > 0) {
 		printf ("selection text \"%.*s\"\n",
 			selection_data->length, selection_data->data); 
 
-		html_engine_delete_selection (GTK_HTML (widget)->engine, TRUE);
-		
 		if (selection_data->type != GDK_SELECTION_TYPE_STRING) {
-			html_engine_insert (GTK_HTML (widget)->engine, 
-					    selection_data->data,
-					    (guint) selection_data->length);
+			html_engine_paste_text (GTK_HTML (widget)->engine, 
+						selection_data->data,
+						(guint) selection_data->length);
 		} else {
 			gchar *utf8 = NULL;
 			
@@ -1226,34 +1233,40 @@ selection_received (GtkWidget *widget,
 						       selection_data->data,
 						       (guint) selection_data->length);
 
-			html_engine_insert (GTK_HTML (widget)->engine, 
-					    utf8, strlen (utf8));
+			html_engine_paste_text (GTK_HTML (widget)->engine, 
+						utf8, unicode_strlen (utf8, -1));
 
 			g_free (utf8);
-		} 
+		}
+		return;
 	}
+
+	if (html_engine_get_editable (GTK_HTML (widget)->engine))
+		html_engine_paste (GTK_HTML (widget)->engine);
 }  
 
 gint
-gtk_html_request_paste (GtkWidget *widget, gint type, gint32 time)
+gtk_html_request_paste (GtkHTML *html, gint type, gint32 time)
 {
 	GdkAtom format_atom;
-	char *formats[] = {"UTF8_STRING", "UTF-8", "STRING"};
+	static char *formats[] = {"UTF8_STRING", "UTF-8", "STRING"};
 
 	if (type >= sizeof (formats) / sizeof (formats[0])) {
 		/* we have now tried all the slection types we support */
-		GTK_HTML (widget)->priv->last_selection_type = -1;
+		html->priv->last_selection_type = -1;
+		if (html_engine_get_editable (html->engine))
+			html_engine_paste (html->engine);
 		return FALSE;
 	}
 	
-	GTK_HTML (widget)->priv->last_selection_type = type;
-	format_atom = gdk_atom_intern (formats[type], FALSE);
+	html->priv->last_selection_type = type;
+	format_atom = gdk_atom_intern (formats [type], FALSE);
 	
 	if (format_atom == GDK_NONE) {
 		g_warning("Could not get requested atom\n");
 	}
 	/* And request the format target for the primary selection */
-	gtk_selection_convert (widget, GDK_SELECTION_PRIMARY, format_atom,
+	gtk_selection_convert (GTK_WIDGET (html), GDK_SELECTION_PRIMARY, format_atom,
 			       time);
 	return TRUE;
 }
@@ -1300,8 +1313,8 @@ set_fonts_idle (GtkHTML *html)
 
 	if (html->engine) {
 		html_font_manager_set_default (&html->engine->painter->font_manager,
-					       prop->font_var_family, prop->font_fix_family,
-					       prop->font_var_size,   prop->font_fix_size);
+					       prop->font_var,      prop->font_fix,
+					       prop->font_var_size, prop->font_fix_size);
 
 		/* tables don't resize correctly :( who knows the solution? */
 		if (html->engine->clue) {
@@ -1770,6 +1783,15 @@ gtk_html_get_type (void)
 	return html_type;
 }
 
+/**
+ * gtk_html_new:
+ * @void:
+ *
+ * GtkHTML widget contructor. Creates an empty GtkHTML widget.
+ *
+ * Return value: newly created empty GtkHTML widget.
+ **/
+
 GtkWidget *
 gtk_html_new (void)
 {
@@ -1839,7 +1861,7 @@ gtk_html_allow_selection (GtkHTML *html,
 
 
 
-/*#define LOG_INPUT*/
+/* #define LOG_INPUT */
 #ifdef LOG_INPUT
 static FILE *log_file;
 #endif
@@ -1860,7 +1882,7 @@ gtk_html_begin_content (GtkHTML *html, gchar *content_type)
 #endif
 	g_return_val_if_fail (! gtk_html_get_editable (html), NULL);
 
-	handle = html_engine_begin (html->engine, html->priv->content_type);
+	handle = html_engine_begin (html->engine, content_type);
 	if (handle == NULL)
 		return NULL;
 
@@ -2047,13 +2069,13 @@ gtk_html_print (GtkHTML *html,
 void
 gtk_html_print_with_header_footer (GtkHTML *html, GnomePrintContext *print_context,
 				   gdouble header_height, gdouble footer_height,
-				   GtkHTMLPrintCallback header_print, GtkHTMLPrintCallback footer_print)
+				   GtkHTMLPrintCallback header_print, GtkHTMLPrintCallback footer_print, gpointer user_data)
 {
 	g_return_if_fail (html != NULL);
 	g_return_if_fail (GTK_IS_HTML (html));
 
 	html_engine_print_with_header_footer (html->engine, print_context,
-					      header_height, footer_height, header_print, footer_print);
+					      header_height, footer_height, header_print, footer_print, user_data);
 }
 
 
@@ -2133,6 +2155,17 @@ gtk_html_set_font_style (GtkHTML *html,
 				 html->engine->insertion_font_style);
 }
 
+void
+gtk_html_set_color (GtkHTML *html, HTMLColor *color)
+{
+	g_return_if_fail (html != NULL);
+	g_return_if_fail (GTK_IS_HTML (html));
+
+	if (html_engine_set_color (html->engine, color))
+		gtk_signal_emit (GTK_OBJECT (html), signals [INSERTION_COLOR_CHANGED],
+				 html->engine->insertion_font_style);
+}
+
 void
 gtk_html_toggle_font_style (GtkHTML *html,
 			    GtkHTMLFontStyle style)
@@ -2175,7 +2208,8 @@ gtk_html_cut (GtkHTML *html)
 	g_return_if_fail (html != NULL);
 	g_return_if_fail (GTK_IS_HTML (html));
 
-	html_engine_cut (html->engine, TRUE);
+	html_engine_cut (html->engine);
+	gtk_selection_owner_set (GTK_WIDGET (html), GDK_SELECTION_PRIMARY, GDK_CURRENT_TIME);
 }
 
 void
@@ -2185,6 +2219,7 @@ gtk_html_copy (GtkHTML *html)
 	g_return_if_fail (GTK_IS_HTML (html));
 
 	html_engine_copy (html->engine);
+	gtk_selection_owner_set (GTK_WIDGET (html), GDK_SELECTION_PRIMARY, GDK_CURRENT_TIME);
 }
 
 void
@@ -2193,7 +2228,7 @@ gtk_html_paste (GtkHTML *html)
 	g_return_if_fail (html != NULL);
 	g_return_if_fail (GTK_IS_HTML (html));
 
-	html_engine_paste (html->engine, TRUE);
+	gtk_html_request_paste (html, 0, GDK_CURRENT_TIME);
 }
 
 
@@ -2261,7 +2296,7 @@ scroll (GtkHTML *html,
 	gfloat         position)
 {
 	GtkAdjustment *adj;
-	gint line_height;
+	gint line25_height;
 	gfloat delta;
 
 	/* we dont want scroll in editable (move cursor instead) */
@@ -2272,8 +2307,8 @@ scroll (GtkHTML *html,
 		? gtk_layout_get_vadjustment (GTK_LAYOUT (html)) : gtk_layout_get_hadjustment (GTK_LAYOUT (html));
 
 
-	line_height = (html->engine && adj->page_increment > html->engine->painter->font_manager.var_size)
-		? html->engine->painter->font_manager.var_size
+	line25_height = (html->engine && adj->page_increment > ((5 * html->engine->painter->font_manager.var_size) >> 1))
+		? ((5 * html->engine->painter->font_manager.var_size) >> 1)
 		: 0;
 
 	switch (scroll_type) {
@@ -2284,10 +2319,10 @@ scroll (GtkHTML *html,
 		delta = -adj->step_increment;
 		break;
 	case GTK_SCROLL_PAGE_FORWARD:
-		delta = adj->page_increment - line_height;
+		delta = adj->page_increment - line25_height;
 		break;
 	case GTK_SCROLL_PAGE_BACKWARD:
-		delta = -adj->page_increment + line_height;
+		delta = -adj->page_increment + line25_height;
 		break;
 	default:
 		g_warning ("invalid scroll parameters: %d %d %f\n", orientation, scroll_type, position);
@@ -2365,22 +2400,33 @@ cursor_move (GtkHTML *html, GtkDirectionType dir_type, GtkHTMLCursorSkipType ski
 			g_warning ("invalid cursor_move parameters\n");
 		}
 		break;
-	case GTK_HTML_CURSOR_SKIP_PAGE:
+	case GTK_HTML_CURSOR_SKIP_PAGE: {
+		gint line25_height;
+
+		line25_height =  GTK_WIDGET (html)->allocation.height
+			> ((5 * html->engine->painter->font_manager.var_size) >> 1)
+			? ((5 * html->engine->painter->font_manager.var_size) >> 1)
+			: 0;
+
+
 		switch (dir_type) {
 		case GTK_DIR_UP:
 		case GTK_DIR_LEFT:
-			if ((amount = html_engine_scroll_up (html->engine, GTK_WIDGET (html)->allocation.height)) > 0)
+			if ((amount = html_engine_scroll_up (html->engine,
+							     GTK_WIDGET (html)->allocation.height - line25_height)) > 0)
 				scroll_by_amount (html, - amount);
 			break;
 		case GTK_DIR_DOWN:
 		case GTK_DIR_RIGHT:
-			if ((amount = html_engine_scroll_down (html->engine, GTK_WIDGET (html)->allocation.height)) > 0)
+			if ((amount = html_engine_scroll_down (html->engine,
+							       GTK_WIDGET (html)->allocation.height - line25_height)) > 0)
 				scroll_by_amount (html, amount);
 			break;
 		default:
 			g_warning ("invalid cursor_move parameters\n");
 		}
 		break;
+	}
 	case GTK_HTML_CURSOR_SKIP_ALL:
 		switch (dir_type) {
 		case GTK_DIR_LEFT:
@@ -2404,6 +2450,7 @@ cursor_move (GtkHTML *html, GtkDirectionType dir_type, GtkHTMLCursorSkipType ski
 	}
 
 	html->binding_handled = TRUE;
+	gtk_html_edit_make_cursor_visible (html);
 }
 
 static void
@@ -2455,6 +2502,7 @@ move_selection (GtkHTML *html, GtkHTMLCommandType com_type)
 	}
 
 	html->binding_handled = TRUE;
+	gtk_html_edit_make_cursor_visible (html);
 }
 
 static void
@@ -2497,40 +2545,44 @@ command (GtkHTML *html, GtkHTMLCommandType com_type)
 		html_engine_redo (e);
 		break;
 	case GTK_HTML_COMMAND_COPY:
-		if (html_engine_is_selection_active (e))
-			html_engine_copy (e);
+		gtk_html_copy (html);
 		break;
 	case GTK_HTML_COMMAND_CUT:
-		html_engine_cut (e, TRUE);
+		gtk_html_cut (html);
 		break;
 	case GTK_HTML_COMMAND_CUT_LINE:
-		html_engine_cut_line (e, TRUE);
+		html_engine_cut_line (e);
 		break;
 	case GTK_HTML_COMMAND_PASTE:
-		if (e->cut_buffer)
-			html_engine_paste (e, TRUE);
+		gtk_html_paste (html);
 		break;
 	case GTK_HTML_COMMAND_INSERT_RULE:
-		html_engine_delete_selection (e, TRUE);
 		html_engine_insert_rule (e, 0, 100, 2, TRUE, HTML_HALIGN_LEFT);
 		break;
 	case GTK_HTML_COMMAND_INSERT_PARAGRAPH:
-		html_engine_delete_selection (e, TRUE);
-		html_engine_insert (e, "\n", 1);
+		html_engine_delete (e);
+		html_engine_insert_empty_paragraph (e);
 		break;
 	case GTK_HTML_COMMAND_DELETE:
 		if (e->mark != NULL
 		    && e->mark->position != e->cursor->position)
-			html_engine_delete_selection (e, TRUE);
+			html_engine_delete (e);
 		else
-			html_engine_delete (e, 1, TRUE, FALSE);
+			html_engine_delete_n (e, 1, TRUE);
 		break;
 	case GTK_HTML_COMMAND_DELETE_BACK:
-		if (e->mark != NULL
-		    && e->mark->position != e->cursor->position)
-			html_engine_delete_selection (e, TRUE);
+		if (html_engine_is_selection_active (e))
+			html_engine_delete (e);
 		else
-			html_engine_delete (e, 1, TRUE, TRUE);
+			html_engine_delete_n (e, 1, FALSE);
+		break;
+	case GTK_HTML_COMMAND_DELETE_BACK_OR_INDENT_DEC:
+		if (html_engine_is_selection_active (e))
+			html_engine_delete (e);
+		else if (html_engine_cursor_on_bop (e) && html_engine_get_indent (e) > 0)
+			gtk_html_modify_indent_by_delta (html, -1);
+		else
+			html_engine_delete_n (e, 1, FALSE);
 		break;
 	case GTK_HTML_COMMAND_SET_MARK:
 		html_engine_set_mark (e);
@@ -2619,6 +2671,9 @@ command (GtkHTML *html, GtkHTMLCommandType com_type)
 	case GTK_HTML_COMMAND_INDENT_DEC:
 		gtk_html_modify_indent_by_delta (html, -1);
 		break;
+	case GTK_HTML_COMMAND_INDENT_PARAGRAPH:
+		html_engine_indent_pre_paragraph (e);
+		break;
 	case GTK_HTML_COMMAND_PARAGRAPH_STYLE_NORMAL:
 		gtk_html_set_paragraph_style (html, GTK_HTML_PARAGRAPH_STYLE_NORMAL);
 		break;
@@ -2790,6 +2845,7 @@ load_keybindings (GtkHTMLClass *klass)
 	/* FIXME add to gtk gtk_binding_set_clear & gtk_binding_set_remove_path */
 	clean_bindings_set (gtk_binding_set_by_class (klass));
 	clean_bindings_set (gtk_binding_set_find ("gtkhtml-bindings-emacs"));
+	clean_bindings_set (gtk_binding_set_find ("gtkhtml-bindings-xemacs"));
 	clean_bindings_set (gtk_binding_set_find ("gtkhtml-bindings-ms"));
 	clean_bindings_set (gtk_binding_set_find ("gtkhtml-bindings-custom"));
 
@@ -2798,6 +2854,7 @@ load_keybindings (GtkHTMLClass *klass)
 	gtk_html_command_get_type ();
 
 	load_bindings_from_file (TRUE,  "keybindingsrc.emacs");
+	load_bindings_from_file (TRUE,  "keybindingsrc.xemacs");
 	load_bindings_from_file (TRUE,  "keybindingsrc.ms");
 	load_bindings_from_file (FALSE, "gtkhtml-bindings-custom");
 
@@ -2873,8 +2930,8 @@ load_keybindings (GtkHTMLClass *klass)
 
 	BCOM (0, Return, INSERT_PARAGRAPH);
 	BCOM (0, KP_Enter, INSERT_PARAGRAPH);
-	BCOM (0, BackSpace, DELETE_BACK);
-	BCOM (GDK_SHIFT_MASK, BackSpace, DELETE_BACK);
+	BCOM (0, BackSpace, DELETE_BACK_OR_INDENT_DEC);
+	BCOM (GDK_SHIFT_MASK, BackSpace, DELETE_BACK_OR_INDENT_DEC);
 	BCOM (0, Delete, DELETE);
 	BCOM (0, KP_Delete, DELETE);
 
@@ -2970,3 +3027,19 @@ gtk_html_editor_command (GtkHTML *html, const gchar *command_name)
 	if (val)
 		command (html, val->value);
 }
+
+gboolean
+gtk_html_edit_make_cursor_visible (GtkHTML *html)
+{
+	gboolean rv = FALSE;
+
+	html_engine_hide_cursor (html->engine);
+	if (html_engine_make_cursor_visible (html->engine)) {
+		gtk_adjustment_set_value (GTK_LAYOUT (html)->hadjustment, (gfloat) html->engine->x_offset);
+		gtk_adjustment_set_value (GTK_LAYOUT (html)->vadjustment, (gfloat) html->engine->y_offset);
+		rv = TRUE;
+	}
+	html_engine_show_cursor (html->engine);
+
+	return rv;
+}
