@@ -147,9 +147,9 @@ void  gtk_html_allow_selection  (GtkHTML  *html,
 				 gboolean  allow);
 void  gtk_html_select_word      (GtkHTML  *html);
 void  gtk_html_select_line      (GtkHTML  *html);
-int   gtk_html_request_paste    (GtkWidget *widget,
-				 gint type,
-				 gint32 time);
+int   gtk_html_request_paste    (GtkHTML  *html,
+				 gint      type,
+				 gint32    time);
 /* Loading.  */
 GtkHTMLStream *gtk_html_begin             (GtkHTML             *html);
 GtkHTMLStream *gtk_html_begin_content     (GtkHTML             *html,
@@ -177,11 +177,6 @@ gboolean  gtk_html_export  (GtkHTML               *html,
 			    GtkHTMLSaveReceiverFn  receiver,
 			    gpointer               data);
 
-/* Streams for feeding the widget with extra data (e.g. images) at loading
-   time.  */
-GtkHTMLStream *gtk_html_stream_ref    (GtkHTMLStream *handle);
-void           gtk_html_stream_unref  (GtkHTMLStream *handle);
-
 /* Editable support.  */
 void      gtk_html_set_editable  (GtkHTML       *html,
 				  gboolean       editable);
@@ -194,7 +189,8 @@ void  gtk_html_print_with_header_footer  (GtkHTML              *html,
 					  gdouble               header_height,
 					  gdouble               footer_height,
 					  GtkHTMLPrintCallback  header_print,
-					  GtkHTMLPrintCallback  footer_print);
+					  GtkHTMLPrintCallback  footer_print,
+					  gpointer              user_data);
 void  gtk_html_print                     (GtkHTML              *html,
 					  GnomePrintContext    *print_context);
 
@@ -218,6 +214,8 @@ void  			   gtk_html_modify_indent_by_delta       (GtkHTML                   *ht
 void  			   gtk_html_set_font_style               (GtkHTML                   *html,
 								  GtkHTMLFontStyle           and_mask,
 								  GtkHTMLFontStyle           or_mask);
+void  			   gtk_html_set_color                    (GtkHTML                   *html,
+								  HTMLColor                 *color);
 void                       gtk_html_toggle_font_style            (GtkHTML                   *html,
 								  GtkHTMLFontStyle           style);
 GtkHTMLParagraphAlignment  gtk_html_get_paragraph_alignment      (GtkHTML                   *html);
@@ -233,10 +231,14 @@ void  gtk_html_redo  (GtkHTML *html);
 
 /* misc utils */
 
-void  gtk_html_set_default_background_color (GtkHTML *html, GdkColor *c);
-void  gtk_html_set_default_content_type (GtkHTML *html, gchar *content_type);
-gpointer  gtk_html_get_object_by_id (GtkHTML *html, const gchar *id);
-
-void  gtk_html_editor_command (GtkHTML *html, const gchar *command_name);
-
+void      gtk_html_set_default_background_color  (GtkHTML     *html,
+						  GdkColor    *c);
+void      gtk_html_set_default_content_type      (GtkHTML     *html,
+						  gchar       *content_type);
+gpointer  gtk_html_get_object_by_id              (GtkHTML     *html,
+						  const gchar *id);
+void      gtk_html_editor_command                (GtkHTML     *html,
+						  const gchar *command_name);
+gboolean  gtk_html_edit_make_cursor_visible      (GtkHTML     *html);
+     
 #endif /* _GTKHTML_H_ */
