@@ -27,7 +27,6 @@
 #include "gtkhtml.h"
 #include "htmlobject.h"
 #include "htmltext.h"
-#include "htmltextmaster.h"
 #include "htmltextslave.h"
 #include "htmltable.h"
 #include "htmlclue.h"
@@ -149,9 +148,7 @@ gtk_html_debug_dump_object (HTMLObject *obj,
 		gtk_html_debug_dump_table (obj, level + 1);
 		break;
 	case HTML_TYPE_TEXT:
-	case HTML_TYPE_TEXTMASTER:
 	case HTML_TYPE_LINKTEXT:
-	case HTML_TYPE_LINKTEXTMASTER:
 		for (i = 0; i < level; i++)
 			g_print (" ");
 		g_print ("Text (%d): \"%s\"\n",
@@ -161,6 +158,7 @@ gtk_html_debug_dump_object (HTMLObject *obj,
 	case HTML_TYPE_CLUEH:
 	case HTML_TYPE_CLUEV:
 	case HTML_TYPE_CLUEFLOW:
+		g_print ("Head: %p Tail: %p\n", HTML_CLUE (obj)->head, HTML_CLUE (obj)->tail);
 	case HTML_TYPE_CLUEALIGNED:
 	case HTML_TYPE_TABLECELL:
 		for (i = 0; i < level; i++) g_print (" ");
@@ -189,6 +187,17 @@ gtk_html_debug_dump_tree (HTMLObject *o,
 }
 
 
+static void
+debug_word_width (HTMLText *t)
+{
+	guint i;
+
+	printf ("words: %d | ", t->words);
+	for (i = 0; i < t->words; i ++)
+		printf ("%d ", t->word_width [i]);
+	printf ("\n");
+}
+
 static void
 dump_object_simple (HTMLObject *obj,
 		    gint level)
@@ -198,17 +207,18 @@ dump_object_simple (HTMLObject *obj,
 	for (i = 0; i < level; i++)
 		g_print ("\t");
 
-	if (html_object_is_text (obj))
+	if (html_object_is_text (obj)) {
 		g_print ("%s `%s'\n",
 			 html_type_name (HTML_OBJECT_TYPE (obj)),
 			 HTML_TEXT (obj)->text);
-	else if (HTML_OBJECT_TYPE (obj) == HTML_TYPE_TEXTSLAVE) {
+		debug_word_width (HTML_TEXT (obj));
+	} else if (HTML_OBJECT_TYPE (obj) == HTML_TYPE_TEXTSLAVE) {
 		HTMLTextSlave *slave = HTML_TEXT_SLAVE (obj);
 		gchar *text;
 
 		text = alloca (slave->posLen+1);
 		text [slave->posLen] = 0;
-		strncpy (text, slave->owner->text.text + slave->posStart, slave->posLen);
+		strncpy (text, slave->owner->text + slave->posStart, slave->posLen);
 		g_print ("%s `%s'\n",
 			 html_type_name (HTML_OBJECT_TYPE (obj)),
 			 text);
