@@ -28,21 +28,23 @@
 #include <config.h>
 #include <ctype.h>
 #include <string.h>
+
 #include "gtkhtml-properties.h"
+
 #include "htmlcolor.h"
 #include "htmlcolorset.h"
 #include "htmlclue.h"
 #include "htmlclueflow.h"
 #include "htmlcluealigned.h"
-#include "htmlpainter.h"
-#include "htmltext.h"
-#include "htmlvspace.h"
-#include "htmllinktextmaster.h"
-#include "htmltextslave.h"	/* FIXME */
-#include "htmlsearch.h"
 #include "htmlentity.h"
 #include "htmlengine-edit.h"
 #include "htmlengine-save.h"
+#include "htmllinktext.h"
+#include "htmlpainter.h"
+#include "htmltext.h"
+#include "htmltextslave.h"	/* FIXME */
+#include "htmlsearch.h"
+#include "htmlvspace.h"
 
 
 HTMLClueFlowClass html_clueflow_class;
@@ -61,6 +63,45 @@ copy (HTMLObject *self,
 	HTML_CLUEFLOW (dest)->level = HTML_CLUEFLOW (self)->level;
 }
 
+static HTMLObject *
+op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
+{
+	if (!from)
+		(*len) ++;
+	html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
+	return (*HTML_OBJECT_CLASS (parent_class)->op_copy) (self, e, from, to, len);
+}
+
+static HTMLObject *
+op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GList *right, guint *len)
+{
+	if (!from)
+		(*len) ++;
+	html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
+	return (*HTML_OBJECT_CLASS (parent_class)->op_cut) (self, e, from, to, left, right, len);
+}
+
+static void
+split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint level, GList **left, GList **right)
+{
+	html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
+	(*HTML_OBJECT_CLASS (parent_class)->split) (self, e, child, offset, level, left, right);
+}
+
+static gboolean
+merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e)
+{
+	HTMLClueFlow *cf1, *cf2;
+
+	cf1 = HTML_CLUEFLOW (self);
+	cf2 = HTML_CLUEFLOW (with);
+
+	html_clueflow_remove_text_slaves (cf1);
+	html_clueflow_remove_text_slaves (cf2);
+
+	return (* HTML_OBJECT_CLASS (parent_class)->merge) (self, with, e);
+}
+
 static guint
 calc_padding (HTMLPainter *painter)
 {
@@ -181,7 +222,7 @@ get_post_padding (HTMLClueFlow *flow,
 		    && ! is_header (next))
 			return pad;
 
-		if (is_header (flow) &&  ! is_header (next))
+		if (is_header (flow))
 			return pad;
 
 		return 0;
@@ -700,10 +741,8 @@ calc_preferred_width (HTMLObject *o,
 			HTMLObject *eol = (obj->flags & HTML_OBJECT_FLAG_NEWLINE) ? html_object_prev_not_slave (obj) : obj;
 
 			/* remove trailing space width on the end of line */
-			if (HTML_OBJECT_TYPE (eol) == HTML_TYPE_TEXTMASTER
-			    || HTML_OBJECT_TYPE (eol) == HTML_TYPE_LINKTEXTMASTER) {
-				w -= html_text_master_trail_space_width (HTML_TEXT_MASTER (eol), painter);
-			}
+			if (html_object_is_text (eol))
+				w -= html_text_trail_space_width (HTML_TEXT (eol), painter);
 
 			if (w > maxw)
 				maxw = w;
@@ -791,8 +830,10 @@ check_point (HTMLObject *self,
 	     guint *offset_return,
 	     gboolean for_cursor)
 {
-	HTMLObject *obj, *p, *pnext;
+	HTMLObject *obj, *p, *pnext, *eol, *cur;
 	HTMLClue *clue;
+	gint line_ly = 0;
+	gint line_y;
 
 	if (x < self->x || x >= self->x + self->width
 	    || y < self->y - self->ascent || y >= self->y + self->descent)
@@ -803,32 +844,77 @@ check_point (HTMLObject *self,
 	x = x - self->x;
 	y = y - self->y + self->ascent;
 
-	for (p = clue->head; p != NULL; p = pnext) {
-		pnext = p->next;
+	for (p = clue->head; p; ) {
 
-		obj = html_object_check_point (p, painter, x, y, offset_return, for_cursor);
+		if (html_object_is_text (p))
+			p = p->next;
+		if (!p)
+			break;
+
+		line_y  = line_ly;
+		line_ly = p->y + p->descent;
+
+		for (eol = p; eol && eol->y - eol->ascent < line_ly; ) {
+			line_ly = MAX (line_ly, eol->y + eol->descent);
+			do
+				eol = eol->next;
+			while (eol && html_object_is_text (eol));
+		}
+
+		if (y >= line_y && y < line_ly)
+			for (cur = p; cur && cur != eol; ) {
+				obj = html_object_check_point (cur, painter, x, for_cursor
+							       ? MIN (MAX (y, cur->y - cur->ascent),
+								      cur->y + cur->descent - 1) : y,
+							       offset_return, for_cursor);
+
+				if (obj != NULL)
+					return obj;
+				do
+					cur = cur->next;
+				while (cur && cur != eol && html_object_is_text (cur));
+			}
+		p = eol;
+	}
+
+	if (!for_cursor)
+		return NULL;
+
+	/* before */
+	p = clue->head;
+	if (html_object_is_text (p))
+		p = p->next;
+	if (x < p->x || y < p->y - p->ascent) {
+		obj = html_object_check_point (p, painter, p->x, p->y - p->ascent, offset_return, for_cursor);
 		if (obj != NULL)
 			return obj;
+	}
+	for (p = clue->head; p != NULL; p = pnext) {
+		if (html_object_is_text (p))
+			p = p->next;
 
-		if (! for_cursor)
-			continue;
+		pnext = p->next;
+		while (pnext && html_object_is_text (pnext))
+			pnext = pnext->next;
 
-		if (p->prev == NULL && (x < p->x || y < p->y - p->ascent)) {
-			obj = html_object_check_point (p, painter, p->x, p->y - p->ascent, offset_return, for_cursor);
-			if (obj != NULL)
-				return obj;
-		} else if (pnext == NULL || (pnext->y != p->y
-					     && y >= p->y - p->ascent
-					     && y <= p->y + p->descent - 1)) {
-			obj = html_object_check_point (HTML_OBJECT_TYPE (p) != HTML_TYPE_TEXTSLAVE
-						       ? p : html_object_prev_not_slave (p),
-						       painter, p->x + p->width - 1, p->y + p->descent -1,
+		/* new line */
+		if (pnext == NULL || (pnext->y - pnext->ascent >= p->y + p->descent
+				      && y >= p->y - p->ascent
+				      && y <  p->y + p->descent)) {
+			obj = html_object_check_point (p, painter, p->x + p->width - 1, p->y + p->descent - 1,
 						       offset_return, for_cursor);
 			if (obj != NULL)
 				return obj;
 		}
 	}
-
+	/* after */
+	p = clue->tail;
+	if (x >= p->x + p->width || y >= p->y + p->descent) {
+		obj = html_object_check_point (p, painter, p->x + p->width - 1, p->y + p->descent - 1,
+					       offset_return, for_cursor);
+		if (obj != NULL)
+			return obj;
+	}
 	return NULL;
 }
 
@@ -1219,8 +1305,7 @@ search_set_info (HTMLObject *cur, HTMLSearch *info, guint pos, guint len)
 	}
 
 	while (cur) {
-		if (HTML_OBJECT_TYPE (cur) == HTML_TYPE_TEXTMASTER
-		    || HTML_OBJECT_TYPE (cur) == HTML_TYPE_LINKTEXTMASTER) {
+		if (html_object_is_text (cur)) {
 			cur_len = HTML_TEXT (cur)->text_len;
 			if (text_len + cur_len > pos) {
 				if (!info->found) {
@@ -1261,8 +1346,7 @@ search_text (HTMLObject **beg, HTMLSearch *info)
 	/* first get flow text_len */
 	text_len = 0;
 	while (cur) {
-		if (HTML_OBJECT_TYPE (cur) == HTML_TYPE_TEXTMASTER
-		    || HTML_OBJECT_TYPE (cur) == HTML_TYPE_LINKTEXTMASTER) {
+		if (html_object_is_text (cur)) {
 			text_len += HTML_TEXT (cur)->text_len;
 			end = cur;
 		} else if (HTML_OBJECT_TYPE (cur) != HTML_TYPE_TEXTSLAVE) {
@@ -1281,8 +1365,7 @@ search_text (HTMLObject **beg, HTMLSearch *info)
 		head = cur = (info->forward) ? *beg : end;
 		cur = *beg;
 		while (cur) {
-			if (HTML_OBJECT_TYPE (cur) == HTML_TYPE_TEXTMASTER
-			    || HTML_OBJECT_TYPE (cur) == HTML_TYPE_LINKTEXTMASTER) {
+			if (html_object_is_text (cur)) {
 				if (!info->forward) {
 					pp -= HTML_TEXT (cur)->text_len;
 				}
@@ -1407,11 +1490,9 @@ search (HTMLObject *obj, HTMLSearch *info)
 		}
 	}
 	while (cur) {
-		if (HTML_OBJECT_TYPE (cur) == HTML_TYPE_TEXTMASTER
-		    || HTML_OBJECT_TYPE (cur) == HTML_TYPE_LINKTEXTMASTER) {
-			if (search_text (&cur, info)) {
+		if (html_object_is_text (cur)) {
+			if (search_text (&cur, info))
 				return TRUE;
-			}
 		} else {
 			html_search_push (info, cur);
 			if (html_object_search (cur, info))
@@ -1478,6 +1559,10 @@ html_clueflow_class_init (HTMLClueFlowClass *klass,
 	/* FIXME destroy */
 
 	object_class->copy = copy;
+	object_class->op_cut = op_cut;
+	object_class->op_copy = op_copy;
+	object_class->split = split;
+	object_class->merge = merge;
 	object_class->calc_size = calc_size;
 	object_class->set_max_width = set_max_width;
 	object_class->calc_min_width = calc_min_width;
@@ -1648,6 +1733,8 @@ html_clueflow_set_style (HTMLClueFlow *flow,
 	g_return_if_fail (engine != NULL);
 	g_return_if_fail (HTML_IS_ENGINE (engine));
 
+	if ((flow->style & HTML_CLUEFLOW_STYLE_PRE) ^ (style & HTML_CLUEFLOW_STYLE_PRE))
+		html_object_clear_word_width (HTML_OBJECT (flow));
 	html_object_change_set (HTML_OBJECT (flow), HTML_CHANGE_ALL);
 	flow->style = style;
 
@@ -1807,12 +1894,12 @@ get_text_bytes (HTMLClue *clue, HTMLInterval *i)
 	g_assert (i->to.object);
 
 	bytes = 0;
-	obj = i->from.object;
-	while (1) {
+	obj = html_interval_get_head (i, HTML_OBJECT (clue));
+	while (obj) {
 		bytes += html_interval_get_bytes (i, obj);
-		if (obj == i->to.object) break;
-		obj = obj->next;
-		g_assert (obj);
+		if (obj == i->to.object)
+			break;
+		obj = html_object_next_not_slave (obj);
 	}
 
 	return bytes;
@@ -1825,11 +1912,11 @@ get_text (HTMLClue *clue, HTMLInterval *i)
 	guint cb, bytes = 0;
 	gchar *text, *ct;
 
-	bytes      = get_text_bytes (clue, i);
-	ct         = text = g_malloc (bytes + 1);
+	bytes        = get_text_bytes (clue, i);
+	ct           = text = g_malloc (bytes + 1);
 	text [bytes] = 0;
 
-	obj = i->from.object;
+	obj = html_interval_get_head (i, HTML_OBJECT (clue));
 	while (obj) {
 		cb = html_interval_get_bytes (i, obj);
 		if (html_object_is_text (obj))
@@ -1838,11 +1925,12 @@ get_text (HTMLClue *clue, HTMLInterval *i)
 			if (cb == 1) *ct = ' ';
 			else memset (ct, ' ', cb);
 		ct += cb;
-		if (obj == i->to.object) break;
-		obj = obj->next;
+		if (obj == i->to.object)
+			break;
+		obj = html_object_next_not_slave (obj);
 	}
 
-	/* printf ("get_text: \"%s\"\n", text); */
+	/* printf ("get_text: %d \"%s\"\n", bytes, text); */
 
 	return text;
 }
@@ -1927,17 +2015,25 @@ end_of_word (gchar *ct)
 	return ct;
 }
 
+static void
+queue_draw (HTMLObject *o, HTMLEngine *e, HTMLInterval *i)
+{
+	if (html_object_is_text (o))
+		html_text_queue_draw (HTML_TEXT (o), e, html_interval_get_start (i, o), html_interval_get_length (i, o));
+}
+
 void
-html_clueflow_spell_check (HTMLClueFlow *flow, HTMLEngine *e, HTMLInterval *i)
+html_clueflow_spell_check (HTMLClueFlow *flow, HTMLEngine *e, HTMLInterval *interval)
 {
 	HTMLObject *obj;
 	HTMLClue *clue;
 	guint off;
 	gchar *text, *ct, *word;
+	HTMLInterval *new_interval = NULL;
 
 	g_assert (HTML_OBJECT_TYPE (flow) == HTML_TYPE_CLUEFLOW);
 
-	/* if (i)
+	/* if (interval)
 	   printf ("html_clueflow_spell_check %p %p %d %d\n", i->from, i->to, i->from_offset, i->to_offset); */
 
 	clue = HTML_CLUE (flow);
@@ -1947,12 +2043,16 @@ html_clueflow_spell_check (HTMLClueFlow *flow, HTMLEngine *e, HTMLInterval *i)
 		return;
 
 	off  = 0;
-	if (!i)
-		i = html_interval_new (clue->head, clue->tail, 0, html_object_get_length (clue->tail));
-	text = get_text (clue, i);
-	obj  = i->from.object;
+
+	if (!interval) {
+		new_interval = html_interval_new (clue->head, clue->tail, 0, html_object_get_length (clue->tail));
+		interval = new_interval;
+	}
+
+	text = get_text (clue, interval);
+	obj  = html_interval_get_head (interval, HTML_OBJECT (flow));
 	if (obj && html_object_is_text (obj))
-		html_text_spell_errors_clear_interval (HTML_TEXT (obj), i);
+		html_text_spell_errors_clear_interval (HTML_TEXT (obj), interval);
 
 	if (text) {
 		ct = text;
@@ -1973,16 +2073,20 @@ html_clueflow_spell_check (HTMLClueFlow *flow, HTMLEngine *e, HTMLInterval *i)
 				if (result == 1) {
 					gboolean is_text = (obj) ? html_object_is_text (obj) : FALSE;
 					while (obj && (!is_text
-						       || (is_text && off + html_interval_get_length (i, obj)
+						       || (is_text && off + html_interval_get_length (interval, obj)
 							   < unicode_index_to_offset (text, ct - text))))
-						obj = next_obj_and_clear (obj, &off, &is_text, i);
+						obj = next_obj_and_clear (obj, &off, &is_text, interval);
 				} else if (obj)
-						obj = spell_check_word_mark (obj, text, word, &off, i);
+						obj = spell_check_word_mark (obj, text, word, &off, interval);
 
 				*ct = bak;
 				if (*ct)
 					ct = unicode_next_utf8 (ct);
 			}
 		}
+		g_free (text);
+		html_interval_forall (interval, e, (HTMLObjectForallFunc) queue_draw, interval);
+		html_engine_flush_draw_queue (e);
+		html_interval_destroy (new_interval);
 	}
 }
