@@ -19,9 +19,11 @@
     Boston, MA 02111-1307, USA.
 */
 
+#include <config.h>
 #include "htmlcluealigned.h"
 #include "htmlcluev.h"
-#include "htmlobject.h"
+#include "htmlengine.h"
+#include "htmlpainter.h"
 
 
 HTMLClueVClass html_cluev_class;
@@ -326,9 +328,9 @@ check_point (HTMLObject *self,
 				x1 = x;
 			}
 
-			if (p->next == NULL && y > p->y + p->descent) {
+			if (p->next == NULL && y > p->y + p->descent - 1) {
 				x1 = p->width - 1;
-				y1 = p->y + p->descent;
+				y1 = p->y + p->descent - 1;
 			} else if (p->prev == NULL && y < p->y - p->ascent) {
 				y1 = p->y - p->ascent;
 			} else {
@@ -532,6 +534,9 @@ find_free_area (HTMLClue *clue, gint y, gint width, gint height,
 		if ((rmargin - lmargin) >= width)
 			break;
 
+		/* if (try_y == next_y)
+		   break; */
+
 		try_y = next_y;
 	}
 	
