@@ -29,6 +29,7 @@
 
 #include "htmlclue.h"
 #include "htmlengine.h"
+#include "htmlengine-edit.h"
 #include "htmltext.h"
 #include "htmltextslave.h"
 #include "htmltype.h"
@@ -167,6 +168,9 @@ html_cursor_home (HTMLCursor *cursor,
 		return;
 	}
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	obj = engine->clue;
 	while (is_clue (obj))
 		obj = HTML_CLUE (obj)->head;
@@ -208,6 +212,9 @@ html_cursor_forward (HTMLCursor *cursor, HTMLEngine *engine)
 	g_return_val_if_fail (cursor != NULL, FALSE);
 	g_return_val_if_fail (engine != NULL, FALSE);
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	cursor->have_target_x = FALSE;
 	retval = forward (cursor);
 
@@ -247,6 +254,9 @@ html_cursor_backward (HTMLCursor *cursor,
 	g_return_val_if_fail (cursor != NULL, FALSE);
 	g_return_val_if_fail (engine != NULL, FALSE);
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	cursor->have_target_x = FALSE;
 	retval = backward (cursor);
 
@@ -274,6 +284,9 @@ html_cursor_up (HTMLCursor *cursor,
 		return TRUE;
 	}
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	html_cursor_copy (&orig_cursor, cursor);
 
 	html_object_get_cursor_base (cursor->object,
@@ -357,6 +370,9 @@ html_cursor_down (HTMLCursor *cursor,
 		return TRUE;
 	}
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	html_object_get_cursor_base (cursor->object,
 				     engine->painter, cursor->offset,
 				     &x, &y);
@@ -440,6 +456,9 @@ html_cursor_jump_to (HTMLCursor *cursor,
 	g_return_val_if_fail (cursor != NULL, FALSE);
 	g_return_val_if_fail (object != NULL, FALSE);
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	html_cursor_normalize (cursor);
 	normalize (&object, &offset);
 
@@ -474,6 +493,9 @@ html_cursor_beginning_of_document (HTMLCursor *cursor,
 	g_return_if_fail (engine != NULL);
 	g_return_if_fail (HTML_IS_ENGINE (engine));
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	while (backward (cursor))
 		;
 }
@@ -486,6 +508,9 @@ html_cursor_end_of_document (HTMLCursor *cursor,
 	g_return_if_fail (engine != NULL);
 	g_return_if_fail (HTML_IS_ENGINE (engine));
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	while (forward (cursor))
 		;
 }
@@ -503,6 +528,9 @@ html_cursor_end_of_line (HTMLCursor *cursor,
 
 	cursor->have_target_x = FALSE;
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	html_cursor_copy (&prev_cursor, cursor);
 	html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
 				     &x, &prev_y);
@@ -536,6 +564,9 @@ html_cursor_beginning_of_line (HTMLCursor *cursor,
 
 	cursor->have_target_x = FALSE;
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	html_cursor_copy (&prev_cursor, cursor);
 	html_object_get_cursor_base (cursor->object, engine->painter, cursor->offset,
 				     &x, &prev_y);
@@ -573,6 +604,9 @@ html_cursor_jump_to_position (HTMLCursor *cursor,
 	g_return_if_fail (cursor != NULL);
 	g_return_if_fail (position >= 0);
 
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
 	if (cursor->position < position) {
 		while (cursor->position < position) {
 			if (! forward (cursor))
@@ -664,3 +698,62 @@ html_cursor_get_prev_char (const HTMLCursor *cursor)
 		? html_text_get_char (HTML_TEXT (prev), HTML_TEXT (prev)->text_len - 1)
 		: 0;
 }
+
+gboolean
+html_cursor_beginning_of_paragraph (HTMLCursor *cursor, HTMLEngine *engine)
+{
+	gboolean rv = FALSE;
+
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
+	while (cursor->offset || html_object_prev_not_slave (cursor->object)) {
+		backward (cursor);
+		rv = TRUE;
+	}
+
+	return rv;
+}
+
+gboolean
+html_cursor_end_of_paragraph (HTMLCursor *cursor, HTMLEngine *engine)
+{
+	gboolean rv = FALSE;
+
+	if (engine->need_spell_check)
+		html_engine_spell_check_range (engine, engine->cursor, engine->cursor);
+
+	while (cursor->offset < html_object_get_length (cursor->object)
+	       || html_object_next_not_slave (cursor->object)) {
+		forward (cursor);
+		rv = TRUE;
+	}
+
+	return rv;
+}
+
+gboolean
+html_cursor_forward_n (HTMLCursor *cursor, HTMLEngine *e, guint n)
+{
+	gboolean rv = FALSE;
+
+	while (n && html_cursor_forward (cursor, e)) {
+		n --;
+		rv = TRUE;
+	}
+
+	return rv;
+}
+
+gboolean
+html_cursor_backward_n (HTMLCursor *cursor, HTMLEngine *e, guint n)
+{
+	gboolean rv = FALSE;
+
+	while (n && html_cursor_backward (cursor, e)) {
+		n --;
+		rv = TRUE;
+	}
+
+	return rv;
+}
