@@ -26,9 +26,13 @@
 #include <ctype.h>
 #include <glib.h>
 
+#include "gtkhtml.h"
+#include "gtkhtml-properties.h"
+
 #include "htmlobject.h"
 #include "htmlclueflow.h"
 #include "htmlcursor.h"
+#include "htmllinktext.h"
 #include "htmltext.h"
 #include "htmltextslave.h"
 #include "htmlimage.h"
@@ -36,11 +40,9 @@
 #include "htmlselection.h"
 #include "htmlundo.h"
 
-#include "htmlengine-cutbuffer.h"
-#include "htmlengine-edit-cut.h"
+#include "htmlengine-edit-cut-and-paste.h"
 #include "htmlengine-edit-cursor.h"
 #include "htmlengine-edit-movement.h"
-#include "htmlengine-edit-paste.h"
 #include "htmlengine-edit-selection-updater.h"
 #include "htmlengine-edit.h"
 
@@ -94,19 +96,22 @@ html_engine_set_mark (HTMLEngine *e)
 }
 
 void
-html_engine_cut_buffer_push (HTMLEngine *e)
+html_engine_clipboard_push (HTMLEngine *e)
 {
-	e->cut_buffer_stack = g_list_prepend (e->cut_buffer_stack, e->cut_buffer);
-	e->cut_buffer = NULL;
+	e->clipboard_stack = g_list_prepend (e->clipboard_stack, GUINT_TO_POINTER (e->clipboard_len));
+	e->clipboard_stack = g_list_prepend (e->clipboard_stack, e->clipboard);
+	e->clipboard       = NULL;
 }
 
 void
-html_engine_cut_buffer_pop (HTMLEngine *e)
+html_engine_clipboard_pop (HTMLEngine *e)
 {
-	g_assert (e->cut_buffer_stack);
+	g_assert (e->clipboard_stack);
 
-	e->cut_buffer = (GList *) e->cut_buffer_stack->data;
-	e->cut_buffer_stack = g_list_remove (e->cut_buffer_stack, e->cut_buffer_stack->data);
+	e->clipboard       = HTML_OBJECT (e->clipboard_stack->data);
+	e->clipboard_stack = g_list_remove (e->clipboard_stack, e->clipboard_stack->data);
+	e->clipboard_len   = GPOINTER_TO_UINT (e->clipboard_stack->data);
+	e->clipboard_stack = g_list_remove (e->clipboard_stack, e->clipboard_stack->data);
 }
 
 void
@@ -155,52 +160,64 @@ html_engine_cut_and_paste_begin (HTMLEngine *e, gchar *op_name)
 {
 	html_engine_hide_cursor (e);
 	html_engine_selection_push (e);
-	html_engine_cut_buffer_push (e);
+	html_engine_clipboard_push (e);
 	html_undo_level_begin (e->undo, op_name);
-	html_engine_cut (e, TRUE);
+	html_engine_cut (e);
 }
 
 void
 html_engine_cut_and_paste_end (HTMLEngine *e)
 {
-	if (e->cut_buffer) {
-		html_engine_paste (e, TRUE);
-		html_engine_cut_buffer_destroy (e->cut_buffer);
+	if (e->clipboard) {
+		html_engine_paste (e);
+		html_engine_clipboard_clear (e);
 	}
 	html_undo_level_end (e->undo);
-	html_engine_cut_buffer_pop (e);
+	html_engine_clipboard_pop (e);
 	html_engine_selection_pop (e);
 	html_engine_show_cursor (e);
 }
 
 void
-html_engine_cut_and_paste (HTMLEngine *e, gchar *op_name, GFunc iterator, gpointer data)
+html_engine_cut_and_paste (HTMLEngine *e, gchar *op_name, HTMLObjectForallFunc iterator, gpointer data)
 {
+	html_engine_edit_selection_updater_update_now (e->selection_updater);
 	html_engine_cut_and_paste_begin (e, op_name);
-	g_list_foreach (e->cut_buffer, iterator, data);
+	if (e->clipboard)
+		html_object_forall (e->clipboard, e, iterator, data);
 	html_engine_cut_and_paste_end (e);
 }
 
+static void
+spell_check_object (HTMLObject *o, HTMLEngine *e, gpointer data)
+{
+	if (HTML_OBJECT_TYPE (o) == HTML_TYPE_CLUEFLOW)
+		html_clueflow_spell_check (HTML_CLUEFLOW (o), e, (HTMLInterval *) data);
+}
+
 void
 html_engine_spell_check_range (HTMLEngine *e, HTMLCursor *begin, HTMLCursor *end)
 {
-	HTMLObject *parent;
 	HTMLInterval *i;
 
+	e->need_spell_check = FALSE;
+
+	if (!e->widget->editor_api || !GTK_HTML_CLASS (GTK_OBJECT (e->widget)->klass)->properties->live_spell_check)
+		return;
+
 	begin = html_cursor_dup (begin);
 	end   = html_cursor_dup (end);
 
-	g_assert (begin->object->parent == end->object->parent);
-
-	parent = begin->object->parent;
-
 	while (html_is_in_word (html_cursor_get_prev_char (begin)))
 		if (html_cursor_backward (begin, e));
 	while (html_is_in_word (html_cursor_get_current_char (end)))
 		if (html_cursor_forward (end, e));
 
 	i = html_interval_new_from_cursor (begin, end);
-	html_clueflow_spell_check (HTML_CLUEFLOW (parent), e, i);
+	if (begin->object->parent != end->object->parent)
+		html_interval_forall (i, e, spell_check_object, i);
+	else
+		html_clueflow_spell_check (HTML_CLUEFLOW (begin->object->parent), e, i);
 	html_interval_destroy (i);
 	html_cursor_destroy (begin);
 	html_cursor_destroy (end);
@@ -263,3 +280,156 @@ html_engine_set_data_by_type (HTMLEngine *e, HTMLType object_type, const gchar *
 
 	g_free (data);
 }
+
+void
+html_engine_clipboard_clear (HTMLEngine *e)
+{
+	if (e->clipboard) {
+		html_object_destroy (e->clipboard);
+		e->clipboard = NULL;
+	}
+}
+
+HTMLObject *
+html_engine_new_text (HTMLEngine *e, const gchar *text, gint len)
+{
+	return e->insertion_url
+		? html_link_text_new_with_len (text, len, e->insertion_font_style, e->insertion_color,
+					       e->insertion_url, e->insertion_target)
+		: html_text_new_with_len      (text, len, e->insertion_font_style, e->insertion_color);
+}
+
+HTMLObject *
+html_engine_new_text_empty (HTMLEngine *e)
+{
+	return html_engine_new_text (e, "", 0);
+}
+
+gboolean
+html_engine_cursor_on_bop (HTMLEngine *e)
+{
+	g_assert (e);
+	g_assert (e->cursor);
+	g_assert (e->cursor->object);
+
+	return e->cursor->offset == 0 && html_object_prev_not_slave (e->cursor->object) == NULL;
+}
+
+guint
+html_engine_get_indent (HTMLEngine *e)
+{
+	g_assert (e);
+	g_assert (e->cursor);
+	g_assert (e->cursor->object);
+
+	return e->cursor->object && e->cursor->object->parent
+		&& HTML_OBJECT_TYPE (e->cursor->object->parent) == HTML_TYPE_CLUEFLOW
+		? HTML_CLUEFLOW (e->cursor->object->parent)->level : 0;
+}
+
+#define LINE_LEN 71
+
+static guint
+try_break_this_line (HTMLEngine *e, guint line_offset, guint last_space)
+{
+	HTMLObject *flow;
+	unicode_char_t uc;
+
+	flow = e->cursor->object->parent;
+
+	while (html_cursor_forward (e->cursor, e) && e->cursor->object->parent == flow) {
+		line_offset ++;
+		uc = html_cursor_get_current_char (e->cursor);
+		if (uc == ' ')
+			last_space = line_offset;
+		if (uc && line_offset >= LINE_LEN) {
+			if (last_space) {
+				html_cursor_backward_n (e->cursor, e, line_offset - last_space);
+				uc = ' ';
+			} else {
+				/* go to end of word */
+				while (html_cursor_forward (e->cursor, e)) {
+					line_offset ++;
+					uc = html_cursor_get_current_char (e->cursor);
+					if (uc == ' ' || !uc)
+						break;
+				}
+			}
+			if (uc == ' ') {
+				html_engine_insert_empty_paragraph (e);
+				html_engine_delete_n (e, 1, TRUE);
+
+				flow        = e->cursor->object->parent;
+				last_space  = 0;
+				line_offset = 0;
+			}
+		}
+		if (!uc)
+			return line_offset;
+	}
+
+	return line_offset;
+}
+
+static void
+go_to_begin_of_pre_para (HTMLEngine *e)
+{
+	HTMLObject *prev;
+
+	do {
+		html_cursor_beginning_of_paragraph (e->cursor, e);
+		prev = html_object_prev_leaf (e->cursor->object);
+		if (prev && html_object_get_length (prev))
+			html_cursor_backward (e->cursor, e);
+		else
+			break;
+	} while (1);
+}
+
+void
+html_engine_indent_pre_paragraph (HTMLEngine *e)
+{
+	guint position;
+	guint line_offset;
+	guint last_space;
+
+	g_assert (e->cursor->object);
+	if (HTML_OBJECT_TYPE (e->cursor->object->parent) != HTML_TYPE_CLUEFLOW
+	    || html_clueflow_get_style (HTML_CLUEFLOW (e->cursor->object->parent)) != HTML_CLUEFLOW_STYLE_PRE)
+		return;
+
+	html_engine_disable_selection (e);
+	position = e->cursor->position;
+
+	html_undo_level_begin (e->undo, "Indent PRE paragraph");
+	html_engine_freeze (e);
+
+	go_to_begin_of_pre_para (e);
+
+	line_offset = 0;
+	last_space  = 0;
+	do {
+		line_offset = try_break_this_line (e, line_offset, last_space);
+		if (html_cursor_forward (e->cursor, e)
+		    && e->cursor->offset == 0 && html_object_get_length (e->cursor->object)
+		    && html_object_prev_not_slave (e->cursor->object) == NULL) {
+			if (line_offset < LINE_LEN - 1) {
+				html_engine_delete_n (e, 1, FALSE);
+				if (' ' != html_cursor_get_prev_char (e->cursor)) {
+					html_engine_insert_text (e, " ", 1);
+					line_offset ++;
+				} else if (position > e->cursor->position)
+					position --;
+				last_space = line_offset - 1;
+			} else {
+				line_offset = 0;
+				last_space  = 0;
+			}
+		} else
+			break;
+	} while (1);
+
+	html_cursor_jump_to_position (e->cursor, e, position);
+	html_engine_thaw (e);
+	html_undo_level_end (e->undo);
+}
