@@ -64,13 +64,16 @@ struct _HTMLEngine {
 	GdkGC *invert_gc;
 
 	gboolean editable;
-	GList *cut_buffer;
-	GList *cut_buffer_stack;
+
+	HTMLObject *clipboard;
+	guint       clipboard_len;
+	GList      *clipboard_stack;
 
 	/* Freeze counter.  When greater than zero, we never trigger relayouts
            nor repaints.  When going from nonzero to zero, we relayout and
            repaint everything.  */
 	guint freeze_count;
+	guint thaw_idle_id;
 
 	gboolean parsing;
 	HTMLTokenizer *ht;
@@ -228,6 +231,8 @@ struct _HTMLEngine {
 	GHashTable *id_table;
 
 	gdouble min_split_index;
+
+	gboolean need_spell_check;
 };
 
 /* must be forward referenced *sigh* */
@@ -288,7 +293,7 @@ void  html_engine_draw_background      (HTMLEngine *e,
 
 /* Scrolling.  */
 void      html_engine_schedule_update      (HTMLEngine  *p);
-void      html_engine_make_cursor_visible  (HTMLEngine  *e);
+gboolean  html_engine_make_cursor_visible  (HTMLEngine  *e);
 gboolean  html_engine_goto_anchor          (HTMLEngine  *e,
 					    const gchar *anchor);
 
@@ -376,4 +381,5 @@ HTMLObject *html_engine_get_object_by_id    (HTMLEngine  *e,
 					     const gchar *id);
 
 HTMLEngine *html_engine_get_top_html_engine (HTMLEngine *e);
+void        html_engine_thaw_idle_reset     (HTMLEngine *e);
 #endif /* _HTMLENGINE_H_ */
