@@ -68,7 +68,7 @@ static void                html_image_animation_destroy (HTMLImageAnimation *ani
 static HTMLImagePointer   *html_image_pointer_new       (const char *filename, HTMLImageFactory *factory);
 static void                html_image_pointer_ref       (HTMLImagePointer *ip);
 static void                html_image_pointer_unref     (HTMLImagePointer *ip);
-
+static gboolean            html_image_pointer_timeout   (HTMLImagePointer *ip);
 static void                render_cur_frame             (HTMLImage *image, gint nx, gint ny, const GdkColor *highlight_color);
 static guint               get_actual_height            (HTMLImage *image, HTMLPainter *painter);
 
@@ -84,7 +84,8 @@ get_actual_width (HTMLImage *image,
 	if (image->percent_width) {
 		/* The cast to `gdouble' is to avoid overflow (eg. when
                    printing).  */
-		width = ((gdouble) html_engine_get_view_width (image->image_ptr->factory->engine)
+		width = ((gdouble) html_engine_get_view_width
+			 (html_engine_get_top_html_engine (image->image_ptr->factory->engine))
 			 * image->specified_width) / 100;
 	} else if (image->specified_width > 0) {
 		width = image->specified_width * html_painter_get_pixel_size (painter);
@@ -120,7 +121,8 @@ get_actual_height (HTMLImage *image,
 	if (image->percent_height) {
 		/* The cast to `gdouble' is to avoid overflow (eg. when
                    printing).  */
-		height = ((gdouble) html_engine_get_view_height (image->image_ptr->factory->engine)
+		height = ((gdouble) html_engine_get_view_height
+			  (html_engine_get_top_html_engine (image->image_ptr->factory->engine))
 			  * image->specified_height) / 100;
 	} else if (image->specified_height > 0) {
 		height = image->specified_height * html_painter_get_pixel_size (painter);
@@ -140,7 +142,7 @@ get_actual_height (HTMLImage *image,
 			height *= scale;
 		} 
 	}
-	
+
 	return height;
 }
 
@@ -236,12 +238,10 @@ calc_preferred_width (HTMLObject *o,
 		      HTMLPainter *painter)
 {
 	HTMLImage *image = HTML_IMAGE (o);
-	guint pixel_size;
 	guint width;
 
-	pixel_size = html_painter_get_pixel_size (painter);
 	width = get_actual_width (HTML_IMAGE (o), painter)
-		+ (image->border * 2 + 2*image->hspace) * pixel_size;
+		+ (image->border * 2 + 2 * image->hspace) * html_painter_get_pixel_size (painter);
 
 	return width;
 }
@@ -313,9 +313,13 @@ draw (HTMLObject *o,
 		hspace = image->hspace * pixel_size;
 		vspace = image->vspace * pixel_size;
 
+		if (image->image_ptr->loader && !image->image_ptr->stall) 
+			return;
+
 		if (o->selected) {
 			html_painter_set_pen (painter, highlight_color);
-			html_painter_fill_rect (painter, o->x + tx + hspace,
+			html_painter_fill_rect (painter, 
+						o->x + tx + hspace,
 						o->y + ty - o->ascent + vspace,
 						o->width - 2 * hspace,
 						o->ascent + o->descent - 2 * vspace);
@@ -329,7 +333,6 @@ draw (HTMLObject *o,
 					 GTK_HTML_ETCH_IN, 1);
 		return;
 	}
-
 	base_x = o->x + tx + (image->border + image->hspace) * pixel_size;
 	base_y = o->y + ty + (image->border + image->vspace) * pixel_size - o->ascent;
 
@@ -766,7 +769,6 @@ html_image_factory_end_pixbuf (GtkHTMLStream *stream,
 	HTMLImagePointer *ip = user_data;
 
 	html_engine_schedule_update (ip->factory->engine);
-	
 	gtk_object_unref (GTK_OBJECT (ip->loader));
 	ip->loader = NULL;
 
@@ -790,18 +792,15 @@ html_image_factory_area_prepared (GdkPixbufLoader *loader, HTMLImagePointer *ip)
 {
 	if (!ip->animation) {
 		GSList *cur;
-		HTMLObject *o;
 
-		ip->pixbuf    = gdk_pixbuf_loader_get_pixbuf (ip->loader);
+		ip->pixbuf = gdk_pixbuf_loader_get_pixbuf (ip->loader);
 		g_assert (ip->pixbuf);
 
 		/* set change flags on images using this image_ptr */
 		cur = ip->interests;
 		while (cur) {
-			if (cur->data) {
-				o = HTML_OBJECT (cur->data);
-				html_object_change_set (o, HTML_CHANGE_ALL);
-			}
+			if (cur->data)
+				html_object_change_set (HTML_OBJECT (cur->data), HTML_CHANGE_ALL);
 			cur = cur->next;
 		}
 
@@ -862,29 +861,6 @@ html_image_animation_timeout (HTMLImage *image)
 
 	frame = (GdkPixbufFrame *) anim->cur_frame->data;
 	
-	/* FIXME - use gdk_pixbuf_composite instead of render_cur_frame
-	   to be more efficient */
-	/* render this step to helper buffer */
-	/*
-	  w = gdk_pixbuf_get_width (gdk_pixbuf_frame_get_pixbuf (frame));
-	  h = gdk_pixbuf_get_height (gdk_pixbuf_frame_get_pixbuf (frame));
-	
-	  if (anim->cur_frame != image->image_ptr->animation->frames) {
-		gdk_pixbuf_composite (gdk_pixbuf_frame_get_pixbuf (frame),
-				      anim->pixbuf,
-				      gdk_pixbuf_frame_get_x_offset (frame),
-				      gdk_pixbuf_frame_get_y_offset (frame),
-				      w, h,
-				      0.0, 0.0, 1.0, 1.0,
-				      ART_FILTER_NEAREST, 255);
-	} else {
-		gdk_pixbuf_copy_area (gdk_pixbuf_frame_get_pixbuf (frame),
-				      0, 0, w, h,
-				      anim->pixbuf,
-				      gdk_pixbuf_frame_get_x_offset (frame),
-				      gdk_pixbuf_frame_get_y_offset (frame));
-				      } */
-
 	/* draw only if animation is active - onscreen */
 
 	engine = image->image_ptr->factory->engine;
@@ -1057,6 +1033,8 @@ html_image_animation_destroy (HTMLImageAnimation *anim)
 	g_free (anim);
 }
 
+#define STALL_INTERVAL 1000
+
 static HTMLImagePointer *
 html_image_pointer_new (const char *filename, HTMLImageFactory *factory)
 {
@@ -1070,10 +1048,41 @@ html_image_pointer_new (const char *filename, HTMLImageFactory *factory)
 	retval->animation = NULL;
 	retval->interests = NULL;
 	retval->factory = factory;
-
+	retval->stall = FALSE;
+	retval->stall_timeout = gtk_timeout_add (STALL_INTERVAL, 
+						 (GtkFunction)html_image_pointer_timeout,
+						 retval);
 	return retval;
 }
 
+static gboolean
+html_image_pointer_timeout (HTMLImagePointer *ip)
+{
+	GSList *list;
+	HTMLImage *image;
+
+	ip->stall = TRUE;
+
+	list = ip->interests;
+	/* 
+	 * draw the frame now that we decided they've had enough time to
+	 * load the image
+	 */
+	if (ip->pixbuf == NULL) {
+		while (list) {
+			image = (HTMLImage *)list->data;
+
+			if (image)
+				html_engine_queue_draw (ip->factory->engine,
+							HTML_OBJECT (image));
+			
+			list = list->next;
+		}
+	}
+	ip->stall_timeout = 0;
+	return FALSE;
+}
+
 static void
 html_image_pointer_ref (HTMLImagePointer *ip)
 {
@@ -1087,6 +1096,10 @@ html_image_pointer_unref (HTMLImagePointer *ip)
 
 	ip->refcount--;
 	if (ip->refcount <= 0) {
+		if (ip->stall_timeout) {
+			gtk_timeout_remove (ip->stall_timeout);
+			ip->stall_timeout = 0;
+		}
 		g_free (ip->url);
 		if (ip->loader) {
 			gtk_object_unref (GTK_OBJECT (ip->loader));
@@ -1097,7 +1110,6 @@ html_image_pointer_unref (HTMLImagePointer *ip)
 		if (ip->pixbuf) {
 			gdk_pixbuf_unref (ip->pixbuf);
 		}
-
 		g_free (ip);
 	}
 }
