@@ -31,10 +31,11 @@
 #include "htmlcolorset.h"
 #include "htmlcursor.h"
 #include "htmlengine.h"
+#include "htmlengine-edit.h"
+#include "htmlinterval.h"
 #include "htmlobject.h"
 #include "htmlpainter.h"
 #include "htmltext.h"
-#include "htmltextmaster.h"
 #include "htmlrule.h"
 #include "htmltype.h"
 
@@ -93,6 +94,86 @@ copy (HTMLObject *self,
 	html_object_copy_data_from_object (dest, self);
 }
 
+static HTMLObject *
+op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
+{
+	gint l = html_object_get_length (self);
+
+	if ((!from || GPOINTER_TO_INT (from->data) == 0) && (!to || GPOINTER_TO_INT (to->data) == l)) {
+		*len += html_object_get_length (self);
+
+		return html_object_dup (self);
+	} else
+		return html_engine_new_text_empty (e);
+}
+
+static HTMLObject *
+op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GList *right, guint *len)
+{
+	gint l = html_object_get_length (self);
+
+	if ((!from || GPOINTER_TO_INT (from->data) == 0) && (!to || GPOINTER_TO_INT (to->data) == l)) {
+		if (!html_object_could_remove_whole (self, from, to, left, right)) {
+			HTMLObject *empty = html_engine_new_text_empty (e);
+
+			if (e->cursor->object == self)
+				e->cursor->object = empty;
+			html_clue_append_after (HTML_CLUE (self->parent), empty, self);
+			html_object_change_set (empty, HTML_CHANGE_ALL);
+			html_object_check_cut_lists (self, empty, left, right);
+		} else
+			html_object_move_cursor_before_remove (self, e);
+
+		html_object_change_set   (self,  HTML_CHANGE_ALL);
+		html_object_remove_child (self->parent, self);
+		*len += l;
+
+
+		return self;
+	} else
+		return html_engine_new_text_empty (e);
+}
+
+static gboolean
+merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e)
+{
+	return FALSE;
+}
+
+static void
+remove_child (HTMLObject *self, HTMLObject *child)
+{
+	g_warning ("REMOVE CHILD unimplemented for ");
+	gtk_html_debug_dump_object_type (self);
+	g_assert_not_reached ();
+}
+
+static void
+split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint level, GList **left, GList **right)
+{
+	if (child || (offset && html_object_get_length (self) != offset)) {
+		g_warning ("don't know how to SPLIT ");
+		gtk_html_debug_dump_object_type (self);
+		return;
+	}
+
+	if (offset) {
+		if (!self->next)
+			html_clue_append (HTML_CLUE (self->parent), html_engine_new_text_empty (e));
+		*left  = g_list_prepend (*left,  self);
+		*right = g_list_prepend (*right, self->next);
+	} else {
+		if (!self->prev)
+			html_clue_prepend (HTML_CLUE (self->parent), html_engine_new_text_empty (e));
+		*left  = g_list_prepend (*left,  self->prev);
+		*right = g_list_prepend (*right, self);
+	}
+	level--;
+
+	if (level && self->parent)
+		html_object_split (self->parent, e, offset ? self->next : self, 0, level, left, right);
+}
+
 static void
 draw (HTMLObject *o,
       HTMLPainter *p,
@@ -407,27 +488,6 @@ select_range (HTMLObject *self,
 	return changed;
 }
 
-static HTMLObject *
-get_selection (HTMLObject *self,
-	       guint *size_return)
-{
-	HTMLObject *copy;
-
-	if (! self->selected) {
-		if (size_return != NULL)
-			*size_return = 0;
-		return NULL;
-	}
-
-	if (size_return != NULL)
-		*size_return = 1;
-
-	copy = html_object_dup (self);
-	html_object_select_range (copy, NULL, 0, 0, FALSE);
-
-	return copy;
-}
-
 static void
 append_selection_string (HTMLObject *self,
 			 GString *buffer)
@@ -532,6 +592,11 @@ html_object_class_init (HTMLObjectClass *klass,
 	/* Install virtual methods.  */
 	klass->destroy = destroy;
 	klass->copy = copy;
+	klass->op_copy = op_copy;
+	klass->op_cut = op_cut;
+	klass->merge = merge;
+	klass->remove_child = remove_child;
+	klass->split = split;
 	klass->draw = draw;
 	klass->draw_background = draw_background;
 	klass->is_transparent = is_transparent;
@@ -559,7 +624,6 @@ html_object_class_init (HTMLObjectClass *klass,
 	klass->get_cursor = get_cursor;
 	klass->get_cursor_base = get_cursor_base;
 	klass->select_range = select_range;
-	klass->get_selection = get_selection;
 	klass->append_selection_string = append_selection_string;
 	klass->forall = forall;
 	klass->is_container = is_container;
@@ -635,6 +699,49 @@ html_object_dup (HTMLObject *object)
 	return new;
 }
 
+HTMLObject *
+html_object_op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
+{
+	return (* HO_CLASS (self)->op_copy) (self, e, from, to, len);
+}
+
+HTMLObject *
+html_object_op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GList *right, guint *len)
+{
+	return (* HO_CLASS (self)->op_cut) (self, e, from, to, left, right, len);
+}
+
+gboolean
+html_object_merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e)
+{
+	if (HTML_OBJECT_TYPE (self) == HTML_OBJECT_TYPE (with)
+	    && (* HO_CLASS (self)->merge) (self, with, e)) {
+		if (with->parent)
+			html_object_remove_child (with->parent, with);
+		html_object_destroy (with);
+		return TRUE;
+	}
+	return FALSE;
+}
+
+void
+html_object_remove_child (HTMLObject *self, HTMLObject *child)
+{
+	g_assert (self);
+	g_assert (child);
+
+	(* HO_CLASS (self)->remove_child) (self, child);
+}
+
+void
+html_object_split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint level,
+		   GList **left, GList **right)
+{
+	g_assert (self);
+
+	(* HO_CLASS (self)->split) (self, e, child, offset, level, left, right);
+}
+
 
 void
 html_object_set_parent (HTMLObject *o, HTMLObject *parent)
@@ -923,13 +1030,6 @@ html_object_select_range (HTMLObject *self,
 	return (* HO_CLASS (self)->select_range) (self, engine, start, length, queue_draw);
 }
 
-HTMLObject *
-html_object_get_selection (HTMLObject *self,
-			   guint *size_return)
-{
-	return (* HO_CLASS (self)->get_selection) (self, size_return);
-}
-
 void
 html_object_append_selection_string (HTMLObject *self,
 				     GString *buffer)
@@ -975,14 +1075,11 @@ html_object_is_text (HTMLObject *object)
 
 	type = HTML_OBJECT_TYPE (object);
 
-	return (type == HTML_TYPE_TEXT
-		|| type == HTML_TYPE_TEXTMASTER
-		|| type == HTML_TYPE_LINKTEXT
-		|| type == HTML_TYPE_LINKTEXTMASTER);
+	return (type == HTML_TYPE_TEXT || type == HTML_TYPE_LINKTEXT);
 }
 
 HTMLObject *
-html_object_next_not_slave (HTMLObject *object)
+html_object_next_not_type (HTMLObject *object, HTMLType t)
 {
 	HTMLObject *p;
 
@@ -990,14 +1087,14 @@ html_object_next_not_slave (HTMLObject *object)
 	g_return_val_if_fail (object->parent, NULL);
 
 	p = html_object_next (object->parent, object);
-	while (p && HTML_OBJECT_TYPE (p) == HTML_TYPE_TEXTSLAVE)
+	while (p && HTML_OBJECT_TYPE (p) == t)
 		p = html_object_next (p->parent, p);
 
 	return p;
 }
 
 HTMLObject *
-html_object_prev_not_slave (HTMLObject *object)
+html_object_prev_not_type (HTMLObject *object, HTMLType t)
 {
 	HTMLObject *p;
 
@@ -1005,12 +1102,24 @@ html_object_prev_not_slave (HTMLObject *object)
 	g_return_val_if_fail (object->parent, NULL);
 
 	p = html_object_prev (object->parent, object);
-	while (p && HTML_OBJECT_TYPE (p) == HTML_TYPE_TEXTSLAVE)
+	while (p && HTML_OBJECT_TYPE (p) == t)
 		p = html_object_prev (p->parent, p);
 
 	return p;
 }
 
+HTMLObject *
+html_object_next_not_slave (HTMLObject *object)
+{
+	return html_object_next_not_type (object, HTML_TYPE_TEXTSLAVE);
+}
+
+HTMLObject *
+html_object_prev_not_slave (HTMLObject *object)
+{
+	return html_object_prev_not_type (object, HTML_TYPE_TEXTSLAVE);
+}
+
 
 gboolean
 html_object_save (HTMLObject *self,
@@ -1151,6 +1260,16 @@ html_object_tail (HTMLObject *self)
 	return (* HO_CLASS (self)->tail) (self);
 }
 
+HTMLObject *
+html_object_tail_not_slave (HTMLObject *self)
+{
+	HTMLObject *o = html_object_tail (self);
+
+	if (o && HTML_OBJECT_TYPE (o) == HTML_TYPE_TEXTSLAVE)
+		o = html_object_prev_not_slave (o);
+	return o;
+}
+
 gboolean
 html_object_cursor_forward (HTMLObject *self, HTMLCursor *cursor)
 {
@@ -1288,3 +1407,136 @@ html_object_copy_data_from_object (HTMLObject *dst, HTMLObject *src)
 {
 	g_datalist_foreach (&src->object_data, copy_data, dst);
 }
+
+GList *
+html_object_get_bound_list (HTMLObject *self, GList *list)
+{
+	return list
+		? (HTML_OBJECT (list->data) == self ? list->next : NULL)
+		: NULL;
+}
+
+void
+html_object_move_cursor_before_remove (HTMLObject *o, HTMLEngine *e)
+{
+	if (e->cursor->object == o) {
+		if (html_object_next_not_slave (o))
+			e->cursor->object = html_object_next_not_slave (o);
+		else
+			e->cursor->object = html_object_prev_not_slave (o);
+	}
+}
+
+gboolean
+html_object_could_remove_whole (HTMLObject *o, GList *from, GList *to, GList *left, GList *right)
+{
+	return ((!from && !to)
+		|| html_object_next_not_slave (HTML_OBJECT (o))
+		|| html_object_prev_not_slave (HTML_OBJECT (o)))
+		&& ((!left  || o != left->data) && (!right || o != right->data));
+
+}
+
+void
+html_object_check_cut_lists (HTMLObject *self, HTMLObject *replacement, GList *left, GList *right)
+{
+	if (left && left->data == self)
+		left->data = replacement;
+	if (right && right->data == self)
+		right->data = replacement;
+}
+
+
+typedef struct {
+	HTMLInterval *i;
+	GString *buffer;
+	gboolean in;
+} tmpSelData;
+
+static void
+select_object (HTMLObject *o, HTMLEngine *e, gpointer data)
+{
+	tmpSelData *d = (tmpSelData *) data;
+
+	if (o == d->i->from.object)
+		d->in = TRUE;
+	if (d->in)
+		html_object_select_range (o, NULL,
+					  html_interval_get_start (d->i, o),
+					  html_interval_get_length (d->i, o), FALSE);
+
+	if (o == d->i->to.object)
+		d->in = FALSE;
+}
+
+static void
+unselect_object (HTMLObject *o, HTMLEngine *e, gpointer data)
+{
+	o->selected = FALSE;
+}
+
+gchar *
+html_object_get_selection_string (HTMLObject *o)
+{
+	HTMLObject *tail;
+	tmpSelData data;
+	gchar *string;
+
+	g_assert (o);
+
+	tail        = html_object_get_tail_leaf (o);
+	data.buffer = g_string_new (NULL);
+	data.in     = FALSE;
+	data.i      = html_interval_new (html_object_get_head_leaf (o), tail, 0, html_object_get_length (tail));
+
+	html_interval_forall (data.i, NULL, select_object, &data);
+	html_object_append_selection_string (o, data.buffer);
+	html_interval_forall (data.i, NULL, unselect_object, NULL);
+
+	html_interval_destroy (data.i);
+	string = data.buffer->str;
+	g_string_free (data.buffer, FALSE);
+
+	return string;
+}
+
+HTMLObject *
+html_object_get_tail_leaf (HTMLObject *o)
+{
+	HTMLObject *tail, *rv = o;
+
+	do {
+		tail = html_object_tail_not_slave (rv);
+		if (tail)
+			rv = tail;
+	} while (tail);
+
+	return rv;
+}
+
+HTMLObject *
+html_object_get_head_leaf (HTMLObject *o)
+{
+	HTMLObject *head, *rv = o;
+
+	do {
+		head = html_object_head (rv);
+		if (head)
+			rv = head;
+	} while (head);
+
+	return rv;
+}
+
+static void
+clear_word_width (HTMLObject *o, HTMLEngine *e, gpointer data)
+{
+	if (html_object_is_text (o))
+		html_text_clear_word_width (HTML_TEXT (o));
+}
+
+void
+html_object_clear_word_width (HTMLObject *o)
+{
+	html_object_forall (o, NULL, clear_word_width, NULL);
+}
